//
//
//  Copyright 2013, Richard Gorton,     All Rights Reserved.
/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.

   The GNU General Public License is contained in the file COPYING.
*/
//
/// \file
/// Definitions for LWP context and ring buffer manipulation
///
/// \author Richard Gorton
///
/// \note:  For this particular implementation, have the constructor
///         allocate both the LWPCB and the Ring Buffer.  Another
///         implementation might choose a different policy.
///
///     The ring buffer can be located anywhere, but the spec recommends
///     the start of a page boundary.

#ifndef _LWPCONTEXT_H_
#define _LWPCONTEXT_H_

#include <limits>
#include <boost/noncopyable.hpp>

#include "lwpcore.h"

/// LWPContext - the combination of an LWPCB and a ring buffer for LWP Version 1
class LWPContext {
    public:

        /// Construct an LWP Context: the LWPCB and the ring buffer.
        /// \param[in]  number of ring buffer events (a minimum of 32) for LWP Version 1
        LWPContext(const size_t& numEvents)
          : m_ringBufferEvents(numEvents),
            m_ringBufferBytes(0),
            m_ringBuffer(NULL),
            m_LWPCB(NULL),
            m_EventConfigBase(NULL)
        {
            // Just in case it hasn't been done already
            LWPInit();

            m_ringBufferEvents = numEvents < (LWPMinBufferUnits() * 32) ? 32 : numEvents;
        }

        ~LWPContext()
        {
            // If we are deleting the LWPCB, we really do need to ensure that LWP is stopped
            // otherwise, unpredictable faults can occur.  Application writer should do this
            // but have one here as well for purposes of defensive coding.
            LWPStop();
            if (NULL != m_LWPCB)
            {
                free(m_LWPCB);
            }
            m_LWPCB = NULL;
        }

        /// Create the ring buffer
        /// \returns    status code of the create.  In this implementation, one possibility is
        ///             LWPStatus_BadRingBufferAction (the producer thread must create the buffer)
        LWP_STATUS  CreateRingBuffer();

        /// Delete the ring buffer - should be performed by the harvester thread
        /// \returns    status code of the deletion.
        ///             LWPStatus_BadRingBufferAction - if the harvester thread was not the deleter
        LWP_STATUS  ReleaseRingBuffer();

        /// Perform all of allocations for the LWPContext.  Since the allocations
        /// might fail, put those in a separate function (best practices)
        /// This does NOT start or customize the LWPCB fields
        /// \returns    status code of the action
        LWP_STATUS  Initialize();

        /// Get the pointer to the LWPCB
        /// \returns    the pointer to the LWPCB
        LWPCB*      GetLWPCB()              { return m_LWPCB; }

        /// the size of the ring buffer (in numbers of events)
        /// \returns    the size of the ring buffer (in numbers of events)
        uint32_t    RingBufferEventSize()   { return m_ringBufferEvents; }

        /// the size of the ring buffer (in bytes)
        /// \returns    the size of the ring buffer (in bytes)
        uint32_t    RingBufferByteSize()    { return m_ringBufferBytes; }

    private:
        size_t              m_ringBufferEvents;
        size_t              m_ringBufferBytes;
        LWP_GenericEvent*   m_ringBuffer;   // Not quite true - the event records in the ring
                                            // buffer vary by type, but are all of the same size
        LWPCB*              m_LWPCB;
        LWPCB_Event*        m_EventConfigBase;  // First event configuration
};

#endif  // _LWPCONTEXT_H_
