//
//
//  Copyright 2012 - 2013, Richard Gorton,     All Rights Reserved.
/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.

   The GNU General Public License is contained in the file COPYING.
*/
/// \file
/// Implementation functions to support the LWP (LightWeight Profiling) extension
/// to the x86_64 architecture.  The LWP specification is desribed
/// in section 13.4 of the AMD64 Architecture Programmer’s Manual Volume 2: System Programming 
/// Caveat: AMD seems to move their web content around a lot, so you should google
/// One link that (presently) works is:
///     http://support.amd.com/us/Embedded_TechDocs/24593.pdf
/// This is a policy neutral (in terms of who manages the LWPCB and Ring Buffers)
/// implementation of a core set of features
///
/// \author Richard Gorton
///
/// \note   This implementation is 64-bit only.
///         Linux: not suppored by default in the kernel
///         Windows8: fully supported (including threshold mode) in the kernel
///         LWP REQUIRES the use of the XSAVE/XRSTOR instructions for context switching
///         See section 13.4.7 of the specification for details.
///
/// Hardware: The "bulldozer" cores support LWP.  An AMD FX-4170 processor implements:
///             LWPVAL, Instructions Retired, Branches Retired, Threshold

#if defined(_WIN32)
#error  Needs development work to support windows (requires Win8)
#endif

#include <stdint.h>
#include <string.h>

#include "lwpcore.h"

LWP_CPUID   g_LWPCharacteristics;

// Must be called prior to usage of other LWP features
bool LWPInit()
{
    static bool initDone = false;

    if (!initDone)
    {
        uint32_t    cpuFeatures[4];

        // Grab all XFeatures capabilities
        __asm__("cpuid"
                : "=a" (cpuFeatures[0]),    // EAX  output operands
                  "=b" (cpuFeatures[1]),    // EBX
                  "=c" (cpuFeatures[2]),    // ECX
                  "=d" (cpuFeatures[3])     // EDX
                : "0" (s_CPUID_XFeaturesSupported),
                  "c" (0) // input operands
                );

        // If LWP, get LWP details
        if (0 != (cpuFeatures[3] & s_CPUID_XFeaturesLWP))
        {
            __asm__("cpuid"
                : "=a" (cpuFeatures[0]),    // EAX  output operands
                  "=b" (cpuFeatures[1]),    // EBX
                  "=c" (cpuFeatures[2]),    // ECX
                  "=d" (cpuFeatures[3])     // EDX
                : "0" (s_CPUID_LWP),
                  "c" (0)       // input operands
                );

            memcpy(&g_LWPCharacteristics, cpuFeatures, sizeof(g_LWPCharacteristics));
        }
        else
        {
            // Clear it completely
            memset(&g_LWPCharacteristics, 0, sizeof(g_LWPCharacteristics));
        }

        initDone = true;
    }

    return LWPAvailable();
}

// Is a particular event available
bool LWPEventAvailable(LWP_EVENTID ev)
{
    bool isAvailable = false;

    if (LWPEventID_Insert_User_Record == ev)
    {
        isAvailable = true;
    }
    else if (0 != ((1 << ev) & g_LWPCharacteristics.OSFeatures))
    {
        isAvailable = true;
    }

    return isAvailable;
}

// Yuk. returning pointers to pointers is nasty, and can be easily screwed up by consumer code.
// So: return a pointer to the event as the return value, and an output status code as a parameter
// Get a pointer to the event, NULL if none exists, and return
// pertinent status
LWPCB_Event* LWPCBEvent(LWPCB* cb, const LWP_EVENTID evid, LWP_STATUS* theStatus)
{
    LWPCB_Event*    eventLocator = NULL;
    LWP_STATUS      status = LWPStatus_Success;
    size_t          offset = 0;
    char*           theLocation;

    if (static_cast<uint32_t>(evid) > LWPMaxEventID())
    {
        status = LWPStatus_EventOutOfRange;
        goto done;
    }
    if (0 == ((1 << evid) & g_LWPCharacteristics.OSFeatures))
    {
        status = LWPStatus_EventNotSupported;
        goto done;
    }

    offset = LWPEventOffset() + ((evid -1) * sizeof(LWPCB_Event));
    theLocation = reinterpret_cast<char*>(cb) + offset;
    eventLocator = reinterpret_cast<LWPCB_Event*>(theLocation);

done: ;
    *theStatus = status;
    return eventLocator;
}

// Configure the paramters of an event
LWP_STATUS LWPConfigureEvent(LWPCB* cb, LWP_EVENTID evid, LWP_EVENTACTION action, uint32_t interval, uint32_t counter)
{
    LWP_STATUS      status = LWPStatus_Success;
    LWPCB_Event*    event;
    uint32_t        theInterval;
    uint32_t        theCount;

    event = LWPCBEvent(cb, evid, &status);
    if (LWPStatus_Success != status)
    {
        goto done;
    }

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    if (LWPEventAction_Enable == action)
    {
        uint32_t    intervalCheck;
        uint32_t    countCheck;

        intervalCheck = interval & s_EventValueMaskVal;
        if (intervalCheck != interval)
        {
            status = LWPStatus_EventIntervalOutOfRange;
            goto done;
        }

        countCheck = counter & s_EventValueMaskVal;
        if (countCheck != counter)
        {
            status = LWPStatus_EventCountOutOfRange;
            goto done;
        }

        theInterval = interval;
        theCount = counter;

        // Finally, set the flags in the LWPCB
        cb->Flags |= (1 << evid);
    }
    else if (LWPEventAction_Disable == action)
    {
        theInterval = 0;
        theCount = 0;

        // CLear the flags in the LWPCB
        cb->Flags &= ~(1 << evid);
    }
    else
    {
        status = LWPStatus_InvalidEventAction;
        goto done;
    }

    event->EventInterval = theInterval;
    event->EventCounter = theCount;

done: ;
    return status;
}

// Enable an event in the LWPCB
LWP_STATUS LWPEnableEvent(LWPCB* cb, LWP_EVENTID evid)
{
    LWPCB_Event*    eventLocator = NULL;
    LWP_STATUS      status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    eventLocator = LWPCBEvent(cb, evid, &status);
    if (LWPStatus_Success != status)
    {
        goto done;
    }
    cb->Flags |= (1 << evid);

done: ;
    return status;
}

// Disable an event in the LWPCB
LWP_STATUS LWPDisableEvent(LWPCB* cb, LWP_EVENTID evid)
{
    LWPCB_Event*    eventLocator = NULL;
    LWP_STATUS      status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    eventLocator = LWPCBEvent(cb, evid, &status);
    if (LWPStatus_Success != status)
    {
        goto done;
    }
    cb->Flags &= ~(1 << evid);

done: ;
    return status;
}

// Set the filter mask
LWP_STATUS LWPSetFilters(LWPCB* cb, uint32_t filters)
{
    LWP_STATUS  status = LWPStatus_Success;
    uint32_t    latencyMask;
    uint32_t    levelFilters;
    uint32_t    branchFilters;
    uint32_t    IPFilters;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    // Be careful not to zap any existing latency bits
    latencyMask = s_AllFiltersMask | 0xff;

    if ((s_AllFiltersMask & filters) != filters)
    {
        status = LWPStatus_InvalidFilter;
        goto done;
    }

    // Make sure the filters are actually supported
    levelFilters = LWP_FILTER_CACHELEVEL
                   | LWP_FILTER_NORTHBRIDGE
                   | LWP_FILTER_REMOTECACHE
                   | LWP_FILTER_RAM
                   | LWP_FILTER_OTHER;
    if (((levelFilters & filters) != 0) && (!LWPCacheLevelFiltering()))
    {
        status = LWPStatus_FilterNotSupported;
        goto done;
    }

    branchFilters = LWP_FILTER_NO_MISPREDICTEDBRANCHES
                    | LWP_FILTER_NO_PREDICTEDBRANCHES
                    | LWP_FILTER_NO_ABSBRANCHES
                    | LWP_FILTER_NO_CONDBRANCHES
                    | LWP_FILTER_NO_RELATIVEBRANCHES;
    if (((branchFilters & filters) != 0) && (!LWPBranchPredictionFiltering()))
    {
        status = LWPStatus_FilterNotSupported;
        goto done;
    }

    IPFilters = LWP_FILTER_IP_INVERT
                | LWP_FILTER_IP_FILTER;
    if (((IPFilters & filters) != 0) && (!LWPIPFiltering()))
    {
        status = LWPStatus_FilterNotSupported;
        goto done;
    }

    cb->Filters = filters;

done: ;
    return status;
}

// Set the latency
LWP_STATUS LWPSetLatency(LWPCB* cb, uint8_t latency)
{
    LWP_STATUS  status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    if (!LWPCacheLatencyFiltering())
    {
        status = LWPStatus_FilterNotSupported;
    }
    else
    {
        cb->Filters &= ~0xff;
        cb->Filters |= latency;
    }
    
done: ;
    return status;
}

/// Set the ring buffer size
/// \param[in]  size (in bytes)
/// \returns    status of the action
LWP_STATUS LWPSetRingBufferByteSize(LWPCB* cb, size_t nbytes)
{
    LWP_STATUS  status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    if (s_MaxRingBufferSize < nbytes)
    {
        status = LWPStatus_RingBufferTooBig;
        goto done;
    }

    cb->BufferSize = nbytes;
done: ;
    return status;
}

// Set the level of randomization (number of bits)
LWP_STATUS LWPSetRandomness(LWPCB* cb, uint32_t rand)
{
    LWP_STATUS  status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    // up to 15 bits of randomness permitted
    if (rand > 15)
    {
        status = LWPStatus_RandomnessTooBig;
        goto done;
    }

    cb->Randomize = rand;

done: ;
    return status;
}

LWP_STATUS LWPSetRingBufferBaseAddress(LWPCB* cb, LWP_GenericEvent* buf)
{
    LWP_STATUS  status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    cb->BufferBaseAddr = reinterpret_cast<uint64_t>(buf);

done: ;
    return status;
}

/// Set the event threshold for interrupts
/// \param[in]  pointer to the LWPCB
/// \param[in]  number of bytes 
/// \returns    status of the action
LWP_STATUS LWPSetThreshold(LWPCB* cb, uint32_t nbytes)
{
    LWP_STATUS  status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    cb->Threshold = nbytes;
done: ;
    return status;
}

/// Set the continuous mode
LWP_STATUS LWPSetContinuousMode(LWPCB* cb, LWP_CMODE mode)
{
    LWP_STATUS  status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    if (!LWPContinuousMode())
    {
        status = LWPStatus_FeatureNotAvailable;
        goto done;
    }

    if (LWPCMode_INVALID == mode)
    {
        status = LWPStatus_UnitializedVariable;
        goto done;
    }

    if (LWPCMode_Continue == mode)
    {
        cb->Flags |= LWP_FLAG_CONTINUOUS;
    }
    else
    {
        cb->Flags &= ~LWP_FLAG_CONTINUOUS;
    }
        

done: ;
    return status;
}

// Set the PTSC mode
LWP_STATUS LWPSetPTSCMode(LWPCB* cb, LWP_PTSCMODE mode)
{
    LWP_STATUS  status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    if (!LWPSavePTSC())
    {
        status = LWPStatus_FeatureNotAvailable;
        goto done;
    }

    if (LWPPTSCMode_INVALID == mode)
    {
        status = LWPStatus_UnitializedVariable;
        goto done;
    }

    if (LWPPTSCMode_Store == mode)
    {
        cb->Flags |= LWP_FLAG_SAVE_PTSC;
    }
    else
    {
        cb->Flags &= ~LWP_FLAG_SAVE_PTSC;
    }

done: ;
    return status;
}

// Set the Threshold Interrupt mode
LWP_STATUS LWPSetThresholdMode(LWPCB* cb, LWP_EVENTACTION action)
{
    LWP_STATUS  status = LWPStatus_Success;

    if (LWPContextIsActive(cb))
    {
        status = LWPStatus_ContextRequiresStop;
        goto done;
    }

    if (!LWPThresholdInterrupts())
    {
        status = LWPStatus_FeatureNotAvailable;
        goto done;
    }

    if (LWPEventAction_INVALID == action)
    {
        status = LWPStatus_UnitializedVariable;
        goto done;
    }

    if (LWPEventAction_Enable == action)
    {
        cb->Flags |= LWP_FLAG_THRESHOLD_INT;
    }
    else
    {
        cb->Flags &= ~LWP_FLAG_THRESHOLD_INT;
    }

done: ;
    return status;
}

// Fetch the number of records currently in the ring buffer
uint32_t RecordsInRingBuffer(LWPCB* cb)
{
    uint32_t    head = cb->BufferHeadOffset;
    uint32_t    tail = cb->BufferTailOffset;
    uint32_t    span = 0;
    uint32_t    nRecs = 0;

    // Account for wrapping
    if (head >= tail)
    {
        // No wrap
        span = head - tail;
    }
    else
    {
        // Wrapped: part at beginning, and part at the end.
        // beginning == head
        // end = size - tail
        span = head + (cb->BufferSize - tail);
    }

    // For V1, the size is going to be 32 bytes.  AKA right shift by 5
    // Leave the general purpose version here (commented out though)
#if 0
    nRecs = span / cb->EventSize;
#else
    nRecs = span >> 5;
#endif

    return nRecs;
}

