//
//
//  Copyright 2013, Richard Gorton,     All Rights Reserved.
/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.

   The GNU General Public License is contained in the file COPYING.
*/
/// \file
/// Two routines:
///     InjectLWP - Sets up the context, and collects instructions retired events
///     FinishLWP - reports content in the ring buffer, cleans up
///

#include <iostream>
#include <ostream>

#include "iretcontext.h"


using namespace std;

extern "C" {
    int     InjectLWP();
    void    FinishLWP();
}

IRetContext*    g_theContext = NULL;

int  InjectLWP()
{
    LWP_STATUS  status = LWPStatus_Success;
    size_t      numEvents;
    uint32_t    evInterval;
    uint32_t    randomize;

    numEvents = 1024*1024;
    evInterval = 5000;

    g_theContext = new IRetContext(numEvents);
    status = g_theContext->Initialize();
    if (LWPStatus_Success != status)
    {
        // Might be the case that the OS has not enabled it.
        if (LWPStatus_LWPNoHardware == status)
        {
            cerr << "LWP not supported in hardware - featureset: "
                 << hex << g_LWPCharacteristics.HardwareFeatures
                 << endl;
        
        }
        if (LWPStatus_LWPNotEnabledbyOS == status)
        {
            cerr << "LWP not enabled in the Operating System\n";
        }

        delete g_theContext;
        return 0;
    }

    // Set up events and filters
    // For purposes of initial experimentation, the setup will be:
    //      Events:         Instructions Retired
    //      Randomization:  6 bits
    //      Threshold:      not enabled (requires OS support)
    //      Filters:        none
    //      Event Interval: 50000

    status = LWPConfigureEvent(g_theContext->GetLWPCB(),
                               LWPEventID_InsnRetired,
                               LWPEventAction_Enable,
                               evInterval,
                               evInterval);     // initial value
    if (LWPStatus_Success != status)
    {
        delete g_theContext;
        return 0;
    }

    randomize = 6;
    status = LWPSetRandomness(g_theContext->GetLWPCB(), randomize);
    if (LWPStatus_Success != status)
    {
        delete g_theContext;
        return 0;
    }

    // Start LWP
    LWPStart(g_theContext->GetLWPCB());

    return 1;
}

// reporting the results for further (if any) consumption
// In this case, some simplistic analysis
static void ReportResults()
{
    float       eventCount;
    uint64_t    uniqueAddrs = 0;
    float       percentUnique;
    uint64_t    totalEvents = 0;

    eventCount = g_theContext->InsnRetiredEvents();

    cout << "Instructions retired report:"
         << "\n\tNumber Instruction Retired Event Records:\t"
         << g_theContext->InsnRetiredEvents()
         << endl;

    // Dump the map - might be worthwhile using a boost::bimap, and
    // walk it in frequency order

    cout << "Histogram of Instructions Retired Addresses:\n\tAddress\tFrequency"
         << endl;

    EventMap_T::const_iterator iter;
    for (iter = g_theContext->IRetEventMap().begin(); iter != g_theContext->IRetEventMap().end(); iter++)
    {
        uniqueAddrs++;
        totalEvents += iter->second;
        cout << "\t"
             << hex << iter->first
             << "\t"
             << dec << iter->second
             << endl;
    }

    percentUnique = static_cast<float>(uniqueAddrs) / eventCount;
    cout << "\nNumber unique addresses: "
         << uniqueAddrs
         << " ("
         << percentUnique
         << "%)"
         << endl;
}

static void ProcessInsnRet(const LWP_InsnsRetired* theInsn)
{
    EventAddress_T  evAddr = reinterpret_cast<EventAddress_T>(theInsn->EffectiveAddr);

    if (0x0 != evAddr)
    {
        g_theContext->IncInsnRetiredEvents(1);
        g_theContext->SetMapEntry(evAddr, 1);
    }
}

void FinishLWP()
{
    LWP_InsnsRetired*   theBuffer;
    uint32_t            ringBufferEvents;

    if (!LWPAvailable())
    {
        // Never allocated a context - we are done
        return;
    }

    // Stop LWP prior to final cleanup.  Very interesting behavior results
    // otherwise - the delete of g_theContext causes the ring buffer to go away
    // which results in a fault during the destructor of the event address map.

    LWPStop();

    ringBufferEvents = g_theContext->RingBufferEventSize();
    theBuffer = reinterpret_cast<LWP_InsnsRetired*>(LWPContextBufferBase(g_theContext->GetLWPCB()));

    for (size_t i = 0; i < ringBufferEvents; i++)
    {
        ProcessInsnRet(&theBuffer[i]);
    }
    ReportResults();
    g_theContext->ReleaseRingBuffer();

    delete g_theContext;
    g_theContext = NULL;
}

